@@ -68,7 +68,7 @@ class ServerEvents(object):
 		'''Override to implement custom handling of the method result and request'''
 		return result
 
-	def log(self, response, txrequest):
+	def log(self, response, txrequest, error=False):
 		'''Override to implement custom error handling'''
 		pass
 
@@ -107,10 +107,7 @@ class JSON_RPC(Resource):
 
 			content = self.eventhandler.processcontent(content, request)
 
-			if isinstance(content, list):
-				content = jsonrpc.common.Request.from_list(content)
-			else:
-				content = jsonrpc.common.Request.from_dict(content)
+			content = jsonrpc.common.Request.from_json(content)
 
 			try:
 				if hasattr(content, 'check'):
@@ -121,9 +118,10 @@ class JSON_RPC(Resource):
 			except jsonrpc.common.RPCError, e:
 				self._ebRender(e, request, content.id if hasattr(content, 'id') else None)
 
-			d = threads.deferToThread(self._action, request, content)
-			d.addCallback(self._cbRender, request)
-			d.addErrback(self._ebRender, request, content.id if hasattr(content, 'id') else None)
+			else:
+				d = threads.deferToThread(self._action, request, content)
+				d.addCallback(self._cbRender, request)
+				d.addErrback(self._ebRender, request, content.id if hasattr(content, 'id') else None)
 		except BaseException, e:
 			self._ebRender(e, request, None)
 
@@ -157,33 +155,39 @@ class JSON_RPC(Resource):
 			if not islist: result = result[0]
 		else: result = None
 
-		self.eventhandler.log(result, request)
 		return result
 
 
+	def _setresponseCode(self, result, request):
+		code = 200
+		if not isinstance(result, list):
+			if result is not None and result.error is not None:
+				code = 500
+		request.setResponseCode(code)
 
 	def _cbRender(self, result, request):
+		self._setresponseCode(result, request)
+		self.eventhandler.log(result, request, error=False)
 		if result is not None:
 			request.setHeader("content-type", 'application/json')
-			request.setResponseCode(200)
 			result = jsonrpc.jsonutil.encode(result).encode('utf-8')
 			request.setHeader("content-length", len(result))
 			request.write(result)
 		request.finish()
 
 	def _ebRender(self, result, request, id, finish=True):
-
 		err = None
 		if not isinstance(result, BaseException):
 			try: result.raiseException()
 			except BaseException, e:
 				err = e
+				self.eventhandler.log(err, request, error=True)
 		else: err = result
+
 		err = self.render_error(err, id)
-		self.eventhandler.log(err, request)
+		self._setresponseCode(err, request)
 
 		request.setHeader("content-type", 'application/json')
-		request.setResponseCode(200)
 		result = jsonrpc.jsonutil.encode(err).encode('utf-8')
 		request.setHeader("content-length", len(result))
 		request.write(result)
